from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING
import logging

from accqsure.exceptions import SpecificationError
from accqsure.documents import Document
from .sections import ChartSections
from .waypoints import ChartWaypoints


if TYPE_CHECKING:
    from accqsure import AccQsure


@dataclass
class Charts:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(f"/chart/{id_}", "GET", kwargs)
        return Chart.from_api(self.accqsure, resp)

    async def list(
        self,
        document_type_id,
        limit=50,
        start_key=None,
        fetch_all=False,
        **kwargs,
    ):
        if fetch_all:
            resp = await self.accqsure._query_all(
                "/chart",
                "GET",
                {
                    "document_type_id": document_type_id,
                    **kwargs,
                },
            )
            charts = [
                Chart.from_api(self.accqsure, chart)
                for chart in resp.get("results")
            ]
            return charts, resp.get("last_key")
        else:
            resp = await self.accqsure._query(
                "/chart",
                "GET",
                {
                    "document_type_id": document_type_id,
                    "limit": limit,
                    "start_key": start_key,
                    **kwargs,
                },
            )
            charts = [
                Chart.from_api(self.accqsure, chart)
                for chart in resp.get("results")
            ]
            return charts, resp.get("last_key")

    async def create(
        self,
        name,
        document_type_id,
        reference_document_id,
        **kwargs,
    ):

        data = dict(
            name=name,
            document_type_id=document_type_id,
            reference_document_id=reference_document_id,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Chart %s", name)

        resp = await self.accqsure._query("/chart", "POST", None, payload)
        chart = Chart.from_api(self.accqsure, resp)
        logging.info("Created Chart %s with id %s", name, chart.id)

        return chart

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(f"/chart/{id_}", "DELETE", {**kwargs})


@dataclass
class Chart:
    id: str
    name: str
    document_type_id: str
    status: str
    created_at: str
    updated_at: str
    reference_document: Optional[Document] = field(default=None)
    approved_by: Optional[str] = field(default=None)
    last_modified_by: Optional[str] = field(default=None)

    sections: ChartSections = field(
        init=False, repr=False, compare=False, hash=False
    )
    waypoints: ChartWaypoints = field(
        init=False, repr=False, compare=False, hash=False
    )

    @classmethod
    def from_api(cls, accqsure: "AccQsure", data: dict[str, Any]) -> "Chart":
        if not data:
            return None
        entity = cls(
            id=data.get("entity_id"),
            name=data.get("name"),
            status=data.get("status"),
            document_type_id=data.get("document_type_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            reference_document=Document.from_api(
                accqsure=accqsure, data=data.get("reference_document")
            ),
            approved_by=data.get("approved_by"),
            last_modified_by=data.get("last_modified_by"),
        )
        entity.accqsure = accqsure
        entity.sections = ChartSections(entity.accqsure, entity.id)
        entity.waypoints = ChartWaypoints(entity.accqsure, entity.id)
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def remove(self):

        await self.accqsure._query(
            f"/chart/{self.id}",
            "DELETE",
        )

    async def rename(self, name):

        resp = await self.accqsure._query(
            f"/chart/{self.id}",
            "PUT",
            None,
            dict(name=name),
        )
        self.__init__(self.accqsure, **resp)
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/chart/{self.id}",
            "GET",
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))
        return self

    async def _set_asset(self, path, file_name, mime_type, contents):
        return await self.accqsure._query(
            f"/chart/{self.id}/asset/{path}",
            "PUT",
            params={"file_name": file_name},
            data=contents,
            headers={"Content-Type": mime_type},
        )

    async def get_reference_contents(self):
        if not self.reference_document:
            raise SpecificationError(
                "reference_document",
                "Reference document not found for chart",
            )
        document_id = self.reference_document.id
        content_id = self.reference_document.content_id
        if not content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )
        resp = await self.accqsure._query(
            f"/document/{document_id}/asset/{content_id}/manifest.json",
            "GET",
        )
        return resp

    async def get_reference_content_item(self, name):
        if not self.reference_document:
            raise SpecificationError(
                "reference_document",
                "Reference document not found for chart",
            )
        document_id = self.reference_document.id
        content_id = self.reference_document.content_id
        if not content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )
        resp = await self.accqsure._query(
            f"/document/{document_id}/asset/{content_id}/{name}",
            "GET",
        )
        return resp



from __future__ import annotations
from dataclasses import dataclass, field, fields
import logging
from typing import Optional, Any, List, Dict, TYPE_CHECKING
from accqsure.charts.waypoints import ChartWaypoint

if TYPE_CHECKING:
    from accqsure import AccQsure


@dataclass
class ChartElements:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    chart_id: str
    section_id: str

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.section_id}/element/{id_}",
            "GET",
            kwargs,
        )
        return ChartElement.from_api(
            self.accqsure, self.chart_id, self.section_id, resp
        )

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.section_id}/element",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        chart_elements = [
            ChartElement.from_api(
                self.accqsure, self.chart_id, self.section_id, chart_element
            )
            for chart_element in resp.get("results")
        ]
        return chart_elements, resp.get("last_key")

    async def create(
        self,
        order,
        element_type,
        description,
        prompt,
        for_each,
        waypoints=None,
        metadata=None,
        **kwargs,
    ):

        data = dict(
            order=order,
            type=element_type,
            description=description,
            prompt=prompt,
            for_each=for_each,
            waypoints=waypoints,
            metadata=metadata,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Chart Element %s", order)

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.section_id}/element",
            "POST",
            None,
            payload,
        )
        chart_element = ChartElement.from_api(
            self.accqsure, self.chart_id, self.section_id, resp
        )
        logging.info("Created Chart %s with id %s", order, chart_element.id)

        return chart_element

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.section_id}/element/{id_}",
            "DELETE",
            {**kwargs},
        )


@dataclass
class ChartElement:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    chart_id: str
    section_id: str
    id: str
    created_at: str
    updated_at: str
    order: int
    type: str
    description: str
    prompt: str
    for_each: bool
    metadata: Optional[Dict[str, Any]] = field(default=None)

    waypoints: Optional[List[ChartWaypoint]] = field(default=None)

    @classmethod
    def from_api(
        cls,
        accqsure: "AccQsure",
        chart_id: str,
        section_id: str,
        data: dict[str, Any],
    ) -> "ChartElement":
        if not data:
            return None
        return cls(
            accqsure=accqsure,
            chart_id=chart_id,
            section_id=section_id,
            id=data.get("entity_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            order=data.get("order"),
            type=data.get("type"),
            description=data.get("description"),
            prompt=data.get("prompt"),
            for_each=data.get("for_each"),
            metadata=data.get("metadata"),
            waypoints=[
                ChartWaypoint.from_api(
                    accqsure=accqsure, chart_id=chart_id, data=waypoint
                )
                for waypoint in data.get("waypoints") or []
            ],
        )

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.section_id}/element/{self.id}",
            "GET",
        )
        exclude = ["id", "chart_id", "section_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))
        return self



from dataclasses import dataclass, field
from typing import Optional
from os import PathLike
import json
import os
import aiohttp
import logging
import traceback
import math
import asyncio
import io
from pathlib import Path
from importlib.metadata import version

from accqsure.auth import Auth
from accqsure.text import Text
from accqsure.document_types import DocumentTypes
from accqsure.documents import Documents
from accqsure.manifests import Manifests
from accqsure.inspections import Inspections
from accqsure.plots import Plots
from accqsure.charts import Charts
from accqsure.util import Utilities

from accqsure.exceptions import (
    ApiError,
    AccQsureException,
    SpecificationError,
    TaskError,
)


DEFAULT_CONFIG_DIR = "~/.accqsure"
DEFAULT_CREDENTIAL_FILE_NAME = "credentials.json"


@dataclass
class AccQsure(object):
    _version = version("accqsure")
    config_dir: Optional[PathLike[str]] = None
    credentials_file: Optional[PathLike[str]] = None
    key: Optional[str] = None

    auth: Auth = field(init=False, repr=False, compare=False, hash=False)
    text: Text = field(init=False, repr=False, compare=False, hash=False)
    document_types: DocumentTypes = field(
        init=False, repr=False, compare=False, hash=False
    )
    documents: Documents = field(
        init=False, repr=False, compare=False, hash=False
    )
    manifests: Manifests = field(
        init=False, repr=False, compare=False, hash=False
    )
    inspections: Inspections = field(
        init=False, repr=False, compare=False, hash=False
    )
    plots: Plots = field(init=False, repr=False, compare=False, hash=False)
    charts: Charts = field(init=False, repr=False, compare=False, hash=False)
    util: Utilities = field(init=False, repr=False, compare=False, hash=False)

    def __post_init__(self, **kwargs):
        self.config_dir = (
            Path(self.config_dir).expanduser().resolve()
            if self.config_dir
            else Path(
                os.environ.get("ACCQSURE_CONFIG_DIR") or DEFAULT_CONFIG_DIR
            )
            .expanduser()
            .resolve()
        )
        self.credentials_file = (
            Path(self.credentials_file).expanduser().resolve()
            if self.credentials_file
            else Path(
                os.environ.get("ACCQSURE_CREDENTIALS_FILE")
                or f"{self.config_dir}/{DEFAULT_CREDENTIAL_FILE_NAME}"
            )
            .expanduser()
            .resolve()
        )
        self.auth = Auth(
            config_dir=self.config_dir,
            credentials_file=self.credentials_file,
            key=self.key,
        )
        self.text = Text(self)
        self.document_types = DocumentTypes(self)
        self.documents = Documents(self)
        self.manifests = Manifests(self)
        self.inspections = Inspections(self)
        self.plots = Plots(self)
        self.charts = Charts(self)
        self.util = Utilities()

    @property
    def __version__(self) -> str:
        return self._version

    async def _query(self, path, method, params=None, data=None, headers=None):
        try:
            token = await self.auth.get_token()
        except AccQsureException as e:
            raise e
        except Exception as e:
            raise AccQsureException(
                f"Error getting authorization tokens.  Verify configured credentials. Error: {traceback.format_exc()}"
            ) from e
        logging.debug(
            "Call parameters - Path: %s, Method: %s, Params: %s, Body: %s, Headers: %s",
            path,
            method,
            params,
            data,
            headers,
        )
        api_endpoint = token.api_endpoint
        headers = (
            {
                **headers,
                **{
                    "Authorization": f"Bearer {token.access_token}",
                    "User-Agent": f"python-sdk/{self._version}",
                },
            }
            if headers
            else {
                "Authorization": f"Bearer {token.access_token}",
                "User-Agent": f"python-sdk/{self._version}",
            }
        )
        if params:
            if not isinstance(params, dict):
                raise AccQsureException(
                    "Query parameters must be a valid dictionary"
                )
            params = {
                k: (str(v).lower() if isinstance(v, bool) else v)
                for k, v in params.items()
                if v is not None
            }  ## aiohttp doesn't support boolean

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        if headers["Content-Type"] == "application/json" and data is not None:
            if isinstance(data, (dict, list, bool, type(None))):
                # Serialize Python objects to JSON string
                data = json.dumps(data)
            if isinstance(data, str):
                # Encode string to bytes
                data = data.encode("utf-8")
            if isinstance(data, (bytes, bytearray)):
                # Wrap bytes in BytesIO to avoid event loop warning
                data = io.BytesIO(data)
            # If data is io.IOBase (e.g., BytesIO, open file), pass as-is for streaming

        url = f"{api_endpoint}/v1{path}"

        logging.debug(
            "Request - Url: %s, Method: %s, Params: %s, Body: %s, Headers: %s",
            url,
            method,
            params,
            data,
            headers,
        )
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method,
                url,
                data=data,
                headers=headers,
                params=params,
            ) as resp:
                if (resp.status // 100) in [4, 5]:
                    what = await resp.read()
                    content_type = resp.headers.get("content-type", "")
                    resp.close()
                    if content_type == "application/json":
                        raise ApiError(
                            resp.status, json.loads(what.decode("utf8"))
                        )
                    else:
                        raise ApiError(
                            resp.status, {"message": what.decode("utf8")}
                        )
                content_type = resp.headers.get("Content-Type", "").lower()
                if "application/json" in content_type:
                    return await resp.json()
                elif "text" in content_type:
                    return await resp.text()
                else:
                    return await resp.read()

    async def _query_stream(
        self, path, method, params=None, data=None, headers=None
    ):
        try:
            token = await self.auth.get_token()
        except AccQsureException as e:
            raise e
        except Exception as e:
            raise AccQsureException(
                f"Error getting authorization tokens.  Verify configured credentials. Error: {traceback.format_exc()}"
            ) from e
        logging.debug(
            "Call parameters - Path: %s, Method: %s, Params: %s, Body: %s, Headers: %s",
            path,
            method,
            params,
            data,
            headers,
        )
        api_endpoint = token.api_endpoint
        headers = (
            {
                **headers,
                **{
                    "Authorization": f"Bearer {token.access_token}",
                    "User-Agent": f"python-sdk/{self._version}",
                },
            }
            if headers
            else {
                "Authorization": f"Bearer {token.access_token}",
                "User-Agent": f"python-sdk/{self._version}",
            }
        )
        if params:
            if not isinstance(params, dict):
                raise AccQsureException(
                    "Query parameters must be a valid dictionary"
                )
            params = {
                k: (str(v).lower() if isinstance(v, bool) else v)
                for k, v in params.items()
                if v is not None
            }  ## aiohttp doesn't support boolean

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        url = f"{api_endpoint}/v1{path}"

        logging.debug(
            "Request - Url: %s, Method: %s, Params: %s, Body: %s, Headers: %s",
            url,
            method,
            params,
            data,
            headers,
        )
        answer = ""
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method,
                url,
                data=json.dumps(data),
                headers=headers,
                params=params,
            ) as resp:
                if (resp.status // 100) in [4, 5]:
                    what = await resp.read()
                    content_type = resp.headers.get("content-type", "")
                    resp.close()
                    if content_type == "application/json":
                        raise ApiError(
                            resp.status, json.loads(what.decode("utf8"))
                        )
                    else:
                        raise ApiError(
                            resp.status, {"message": what.decode("utf8")}
                        )
                try:
                    async for line in resp.content:
                        if line and line.strip():
                            clean_line = (
                                line.decode("utf-8")
                                .removeprefix("data:")
                                .strip()
                            )
                            # pylint: disable=no-member
                            # logging.trace(clean_line)
                            if clean_line == "[DONE]":
                                continue
                            try:
                                response = json.loads(clean_line)
                            except Exception:
                                logging.error("bad line: %s", clean_line)
                                continue

                            if response.get("generated_text"):
                                # pylint: disable=no-member
                                # logging.trace("final response: %s", response)
                                return response.get("generated_text")
                            elif response.get("choices")[0].get(
                                "finish_reason"
                            ):
                                continue
                            else:
                                content = (
                                    response.get("choices")[0]
                                    .get("delta")
                                    .get("content")
                                )
                                answer += content
                    return answer
                except Exception as e:
                    logging.exception("Error during generation response")
                    data = await response.text()
                    logging.error("Response error: %s", data)
                    raise e

    async def _query_all(
        self, path, method, params=None, data=None, headers=None
    ):
        all_results = []
        params = params or {}  # Ensure params is a dict
        params["limit"] = params.get(
            "limit", 100
        )  # Set default limit if not provided
        cursor = None

        while True:
            # Update params with the current cursor (start_key)
            if cursor:
                params["start_key"] = cursor
            else:
                params.pop("start_key", None)  # Remove start_key if no cursor

            # Make the API call
            resp = await self._query(
                path=path,
                method=method,
                params=params,
                data=data,
                headers=headers,
            )

            # Extract results and cursor
            results = resp.get("results", [])
            all_results.extend(results)
            cursor = resp.get("last_key")

            # Break if no more cursor
            if not cursor:
                break

        return all_results

    async def _poll_task(self, task_id, timeout=300):
        MAX_TIMEOUT = 24 * 60 * 60
        if timeout > MAX_TIMEOUT:
            raise SpecificationError(
                "timeout",
                f"timeout must be less than {MAX_TIMEOUT} seconds.",
            )

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(
            min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN
        )
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)

            resp = await self._query(
                f"/task/{task_id}",
                "GET",
            )

            status = resp.get("status")
            if status == "finished":
                return resp.get("result")
            if status in ["failed", "canceled"]:
                raise TaskError(resp.get("result"))

            count += 1

        raise AccQsureException(f"Timeout waiting for task {task_id}")



from __future__ import annotations
from dataclasses import dataclass, field, fields
import logging
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from accqsure import AccQsure


@dataclass
class ChartWaypoints(object):
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    chart_id: str

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/waypoint/{id_}", "GET", kwargs
        )
        return ChartWaypoint.from_api(self.accqsure, self.chart_id, resp)

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/waypoint",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        chart_waypoints = [
            ChartWaypoint.from_api(
                self.accqsure, self.chart_id, chart_waypoint
            )
            for chart_waypoint in resp.get("results")
        ]
        return chart_waypoints, resp.get("last_key")

    async def create(
        self,
        name,
        **kwargs,
    ):

        data = dict(
            name=name,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Chart Waypoint %s", name)

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/waypoint", "POST", None, payload
        )
        chart_waypoint = ChartWaypoint.from_api(
            self.accqsure, self.chart_id, resp
        )
        logging.info(
            "Created Chart Waypoint %s with id %s", name, chart_waypoint.id
        )

        return chart_waypoint

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(
            f"/chart/{self.chart_id}/waypoint/{id_}", "DELETE", {**kwargs}
        )


@dataclass
class ChartWaypoint:
    chart_id: str
    id: str
    name: str
    created_at: str
    updated_at: str

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", chart_id: str, data: dict[str, Any]
    ) -> "ChartWaypoint":
        if not data:
            return None

        entity = cls(
            chart_id=chart_id,
            id=data.get("entity_id"),
            name=data.get("name"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/waypoint/{self.id}",
            "GET",
        )
        exclude = ["id", "chart_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))

        return self



from __future__ import annotations
from dataclasses import dataclass, field, fields
import logging
from typing import Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from accqsure import AccQsure


from .elements import ChartElements


@dataclass
class ChartSections(object):
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    chart_id: str

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{id_}", "GET", kwargs
        )
        return ChartSection.from_api(self.accqsure, self.chart_id, resp)

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        chart_sections = [
            ChartSection.from_api(self.accqsure, self.chart_id, chart_section)
            for chart_section in resp.get("results")
        ]
        return chart_sections, resp.get("last_key")

    async def create(
        self,
        heading,
        style,
        order,
        number=None,
        **kwargs,
    ):

        data = dict(
            heading=heading,
            style=style,
            order=order,
            number=number,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Chart Section %s", order)

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section", "POST", None, payload
        )
        chart_section = ChartSection.from_api(
            self.accqsure, self.chart_id, resp
        )
        logging.info(
            "Created Chart Section %s with id %s", order, chart_section.id
        )

        return chart_section

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{id_}", "DELETE", {**kwargs}
        )


@dataclass
class ChartSection:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    chart_id: str
    id: str
    created_at: str
    updated_at: str
    heading: str
    style: str
    order: int
    number: Optional[str] = field(default=None)

    elements: ChartElements = field(
        init=False, repr=False, compare=False, hash=False
    )

    def __post_init__(self):
        self.elements = ChartElements(self.accqsure, self.chart_id, self.id)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", chart_id: str, data: dict[str, Any]
    ) -> "ChartSection":
        if not data:
            return None
        return cls(
            accqsure=accqsure,
            chart_id=chart_id,
            id=data.get("entity_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            heading=data.get("heading"),
            number=data.get("number"),
            style=data.get("style"),
            order=data.get("order"),
        )

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/chart/{self.chart_id}/section/{self.id}",
            "GET",
        )
        exclude = ["id", "chart_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))

        return self



import click
import os
import base64
import mimetypes
from enum import Enum
from accqsure.cli import cli, pass_config


@cli.group()
@pass_config
def chart(config):
    """AccQsure chart commands."""
    pass


@chart.command()
@click.argument("document_type_id", type=click.STRING)
@pass_config
def list(config, document_type_id):
    """List charts."""
    data = [
        ["ID", "NAME", "REFERENCE_DOC", "REFERENCE_DOC_ID"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    # TODO: implement cursor function
    charts, _ = config.accqsure.run(
        config.accqsure.client.charts.list(document_type_id=document_type_id)
    )
    for chart in charts:
        data.append(
            [
                chart.id,
                chart.name,
                (
                    chart.reference_document.name
                    if chart.reference_document
                    else "UNKNOWN"
                ),
                (
                    chart.reference_document.id
                    if chart.reference_document
                    else "UNKNOWN"
                ),
            ]
        )
    for row in data:
        click.echo(
            "{: >26.24} {: >28.26} {: >28.26} {: >26.24} " "".format(*row),
            file=config.stdout,
        )



from __future__ import annotations
from dataclasses import dataclass, field, fields
import logging
from typing import Optional, Any, TYPE_CHECKING

from accqsure.exceptions import SpecificationError

if TYPE_CHECKING:
    from accqsure import AccQsure


@dataclass
class Documents:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(f"/document/{id_}", "GET", kwargs)
        return Document.from_api(self.accqsure, resp)

    async def list(self, document_type_id, **kwargs):
        resp = await self.accqsure._query(
            "/document",
            "GET",
            dict(document_type_id=document_type_id, **kwargs),
        )

        documents = [
            Document.from_api(self.accqsure, document)
            for document in resp.get("results")
        ]
        return documents, resp.get("last_key")

    async def create(
        self,
        document_type_id,
        name,
        doc_id,
        contents,
        **kwargs,
    ):

        data = dict(
            name=name,
            document_type_id=document_type_id,
            doc_id=doc_id,
            contents=contents,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Document %s", name)

        resp = await self.accqsure._query("/document", "POST", None, payload)
        document = Document.from_api(self.accqsure, resp)
        logging.info("Created Document %s with id %s", name, document.id)

        return document

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(
            f"/document/{id_}", "DELETE", dict(**kwargs)
        )


@dataclass
class Document:
    accqsure: "AccQsure" = field(repr=False, compare=False, hash=False)
    id: str
    name: str
    status: str
    doc_id: str
    created_at: str
    updated_at: str
    document_type_id: Optional[str] = field(default=None)
    content_id: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", data: dict[str, Any]
    ) -> "Document":
        if not data:
            return None
        return cls(
            accqsure=accqsure,
            id=data.get("entity_id"),
            name=data.get("name"),
            status=data.get("status"),
            document_type_id=data.get("document_type_id"),
            doc_id=data.get("doc_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            content_id=data.get("content_id"),
        )

    async def remove(self):

        await self.accqsure._query(
            f"/document/{self.id}",
            "DELETE",
        )

    async def rename(self, name):

        resp = await self.accqsure._query(
            f"/document/{self.id}",
            "PUT",
            None,
            dict(name=name),
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/document/{self.id}",
            "GET",
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args (skip derived like sections/waypoints)
                setattr(self, f.name, resp.get(f.name))
        return self

    async def get_contents(self):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )

        resp = await self.accqsure._query(
            f"/document/{self.id}/asset/{self.content_id}/manifest.json",
            "GET",
        )
        return resp

    async def get_content_item(self, name):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )

        return await self.accqsure._query(
            f"/document/{self.id}/asset/{self.content_id}/{name}",
            "GET",
        )

    async def _set_asset(self, path, file_name, mime_type, contents):
        return await self.accqsure._query(
            f"/document/{self.id}/asset/{path}",
            "PUT",
            params={"file_name": file_name},
            data=contents,
            headers={"Content-Type": mime_type},
        )

    async def _set_content_item(self, name, file_name, mime_type, contents):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )

        return await self._set_asset(
            f"{self.content_id}/{name}", file_name, mime_type, contents
        )
